@@ -1,39 +1,14 @@
-# -*- coding: utf-8 -*-
-#
-# Copyright (c) 2011-2016, Cédric Krier
-# Copyright (c) 2013-2014, Nicolas Évrard
-# Copyright (c) 2011-2016, B2CK
-# All rights reserved.
-#
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are met:
-#     * Redistributions of source code must retain the above copyright
-#       notice, this list of conditions and the following disclaimer.
-#     * Redistributions in binary form must reproduce the above copyright
-#       notice, this list of conditions and the following disclaimer in the
-#       documentation and/or other materials provided with the distribution.
-#     * Neither the name of the <organization> nor the
-#       names of its contributors may be used to endorse or promote products
-#       derived from this software without specific prior written permission.
-#
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
-# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
-# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
-# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
+# This file is part of python-sql.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import unittest
 import warnings
 from copy import deepcopy
 
-from sql import Table, Join, Union, Literal, Flavor, For, With, Window, Select
-from sql.functions import Now, Function, Rank, DatePart
-from sql.aggregate import Min
+from sql import (
+    Cube, Flavor, For, Grouping, Join, Literal, Rollup, Select, Table, Union,
+    Window, With)
+from sql.aggregate import Max, Min
+from sql.functions import DatePart, Function, Now, Rank
 
 
 class TestSelect(unittest.TestCase):
@@ -42,12 +17,12 @@ class TestSelect(unittest.TestCase):
     def test_select1(self):
         query = self.table.select()
         self.assertEqual(str(query), 'SELECT * FROM "t" AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_select2(self):
         query = self.table.select(self.table.c)
         self.assertEqual(str(query), 'SELECT "a"."c" FROM "t" AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         query.columns += (self.table.c2,)
         self.assertEqual(str(query),
@@ -57,35 +32,41 @@ class TestSelect(unittest.TestCase):
         query = self.table.select(where=(self.table.c == 'foo'))
         self.assertEqual(str(query),
             'SELECT * FROM "t" AS "a" WHERE ("a"."c" = %s)')
-        self.assertEqual(query.params, ('foo',))
+        self.assertEqual(tuple(query.params), ('foo',))
 
     def test_select_without_from(self):
         query = Select([Literal(1)])
         self.assertEqual(str(query), 'SELECT %s')
-        self.assertEqual(query.params, (1,))
+        self.assertEqual(tuple(query.params), (1,))
 
     def test_select_select(self):
         query = Select([Select([Literal(1)])])
         self.assertEqual(str(query), 'SELECT (SELECT %s)')
-        self.assertEqual(query.params, (1,))
+        self.assertEqual(tuple(query.params), (1,))
 
     def test_select_select_as(self):
         query = Select([Select([Literal(1)]).as_('foo')])
         self.assertEqual(str(query), 'SELECT (SELECT %s) AS "foo"')
-        self.assertEqual(query.params, (1,))
+        self.assertEqual(tuple(query.params), (1,))
+
+    def test_select_distinct(self):
+        query = self.table.select(self.table.c, distinct=True)
+        self.assertEqual(
+            str(query), 'SELECT DISTINCT "a"."c" FROM "t" AS "a"')
+        self.assertEqual(tuple(query.params), ())
 
     def test_select_distinct_on(self):
         query = self.table.select(self.table.c, distinct_on=self.table.c)
         self.assertEqual(
             str(query), 'SELECT DISTINCT ON ("a"."c") "a"."c" FROM "t" AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         query = self.table.select(
             self.table.c, distinct_on=[self.table.a, self.table.b])
         self.assertEqual(
             str(query),
             'SELECT DISTINCT ON ("a"."a", "a"."b") "a"."c" FROM "t" AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_select_from_list(self):
         t2 = Table('t2')
@@ -93,7 +74,7 @@ class TestSelect(unittest.TestCase):
         query = (self.table + t2 + t3).select(self.table.c, getattr(t2, '*'))
         self.assertEqual(str(query),
             'SELECT "a"."c", "b".* FROM "t" AS "a", "t2" AS "b", "t3" AS "c"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_select_union(self):
         query1 = self.table.select()
@@ -113,11 +94,11 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(union),
             'SELECT * FROM "t" AS "a" WHERE ("a"."c" = %s) UNION ALL '
             'SELECT * FROM "t2" AS "b"')
-        self.assertEqual(union.params, ('foo',))
+        self.assertEqual(tuple(union.params), ('foo',))
 
         union = Union(query1)
         self.assertEqual(str(union), str(query1))
-        self.assertEqual(union.params, query1.params)
+        self.assertEqual(tuple(union.params), tuple(query1.params))
 
     def test_select_union_order(self):
         query1 = self.table.select()
@@ -128,7 +109,7 @@ class TestSelect(unittest.TestCase):
             'SELECT * FROM "t" AS "a" UNION '
             'SELECT * FROM "t2" AS "b" '
             'ORDER BY %s')
-        self.assertEqual(union.params, (1,))
+        self.assertEqual(tuple(union.params), (1,))
 
     def test_select_intersect(self):
         query1 = self.table.select()
@@ -137,7 +118,7 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(intersect),
             'SELECT * FROM "t" AS "a" INTERSECT SELECT * FROM "t2" AS "b"')
 
-        from sql import Intersect, Interesect
+        from sql import Interesect, Intersect
         with warnings.catch_warnings(record=True) as w:
             warnings.simplefilter('always')
             interesect = Interesect(query1, query2)
@@ -170,12 +151,12 @@ class TestSelect(unittest.TestCase):
         select = t1.select()
         self.assertEqual(str(select.select()),
             'SELECT * FROM (SELECT * FROM "t1" AS "b") AS "a"')
-        self.assertEqual(select.params, ())
+        self.assertEqual(tuple(select.params), ())
 
     def test_select_function(self):
         query = Now().select()
         self.assertEqual(str(query), 'SELECT * FROM NOW() AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_select_function_columns_definitions(self):
         class Crosstab(Function):
@@ -186,25 +167,69 @@ class TestSelect(unittest.TestCase):
                 ('c1', 'INT'), ('c2', 'CHAR'), ('c3', 'BOOL')]).select()
         self.assertEqual(str(query), 'SELECT * FROM CROSSTAB(%s, %s) '
             'AS "a" ("c1" INT, "c2" CHAR, "c3" BOOL)')
-        self.assertEqual(query.params, ('query1', 'query2'))
+        self.assertEqual(tuple(query.params), ('query1', 'query2'))
 
     def test_select_group_by(self):
         column = self.table.c
         query = self.table.select(column, group_by=column)
         self.assertEqual(str(query),
             'SELECT "a"."c" FROM "t" AS "a" GROUP BY "a"."c"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         output = column.as_('c1')
         query = self.table.select(output, group_by=output)
         self.assertEqual(str(query),
             'SELECT "a"."c" AS "c1" FROM "t" AS "a" GROUP BY "c1"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         query = self.table.select(Literal('foo'), group_by=Literal('foo'))
         self.assertEqual(str(query),
             'SELECT %s FROM "t" AS "a" GROUP BY %s')
-        self.assertEqual(query.params, ('foo', 'foo'))
+        self.assertEqual(tuple(query.params), ('foo', 'foo'))
+
+    def test_select_group_by_grouping_sets(self):
+        query = self.table.select(
+            Literal('*'),
+            group_by=Grouping((self.table.a, self.table.b), (Literal('foo'),)))
+        self.assertEqual(str(query),
+            'SELECT %s FROM "t" AS "a" '
+            'GROUP BY GROUPING SETS (("a"."a", "a"."b"), (%s))')
+        self.assertEqual(tuple(query.params), ('*', 'foo',))
+
+        query = self.table.select(
+            Literal('*'),
+            group_by=[
+                self.table.a, Grouping((self.table.b,), (self.table.c,))])
+        self.assertEqual(str(query),
+            'SELECT %s FROM "t" AS "a" '
+            'GROUP BY "a"."a", GROUPING SETS (("a"."b"), ("a"."c"))')
+        self.assertEqual(tuple(query.params), ('*',))
+
+    def test_select_group_by_rollup(self):
+        query = self.table.select(
+            Literal('*'),
+            group_by=Rollup(self.table.a, self.table.b, Literal('foo')))
+        self.assertEqual(str(query),
+            'SELECT %s FROM "t" AS "a" '
+            'GROUP BY ROLLUP ("a"."a", "a"."b", %s)')
+        self.assertEqual(tuple(query.params), ('*', 'foo'))
+
+        query = self.table.select(
+            Literal('*'),
+            group_by=Rollup((self.table.a, self.table.b), self.table.c))
+        self.assertEqual(str(query),
+            'SELECT %s FROM "t" AS "a" '
+            'GROUP BY ROLLUP (("a"."a", "a"."b"), "a"."c")')
+        self.assertEqual(tuple(query.params), ('*',))
+
+    def test_select_group_by_cube(self):
+        query = self.table.select(
+            Literal('*'),
+            group_by=Cube(self.table.a, self.table.b))
+        self.assertEqual(str(query),
+            'SELECT %s FROM "t" AS "a" '
+            'GROUP BY CUBE ("a"."a", "a"."b")')
+        self.assertEqual(tuple(query.params), ('*',))
 
     def test_select_having(self):
         col1 = self.table.col1
@@ -214,14 +239,14 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(query),
             'SELECT "a"."col1", MIN("a"."col2") FROM "t" AS "a" '
             'HAVING (MIN("a"."col2") > %s)')
-        self.assertEqual(query.params, (3,))
+        self.assertEqual(tuple(query.params), (3,))
 
     def test_select_order(self):
         c = self.table.c
         query = self.table.select(c, order_by=Literal(1))
         self.assertEqual(str(query),
             'SELECT "a"."c" FROM "t" AS "a" ORDER BY %s')
-        self.assertEqual(query.params, (1,))
+        self.assertEqual(tuple(query.params), (1,))
 
     def test_select_limit_offset(self):
         try:
@@ -229,34 +254,34 @@ class TestSelect(unittest.TestCase):
             query = self.table.select(limit=50, offset=10)
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a" LIMIT 50 OFFSET 10')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.limit = None
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a" OFFSET 10')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.offset = 0
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a"')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             Flavor.set(Flavor(limitstyle='limit', max_limit=-1))
 
             query.offset = None
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a"')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.offset = 0
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a"')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.offset = 10
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a" LIMIT -1 OFFSET 10')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
         finally:
             Flavor.set(Flavor())
 
@@ -267,17 +292,17 @@ class TestSelect(unittest.TestCase):
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a" '
                 'OFFSET (10) ROWS FETCH FIRST (50) ROWS ONLY')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.limit = None
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a" OFFSET (10) ROWS')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
 
             query.offset = 0
             self.assertEqual(str(query),
                 'SELECT * FROM "t" AS "a"')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
         finally:
             Flavor.set(Flavor())
 
@@ -291,7 +316,7 @@ class TestSelect(unittest.TestCase):
                         'SELECT * FROM "t" AS "c") AS "b" '
                     'WHERE (ROWNUM <= %s)) AS "a" '
                 'WHERE ("rnum" > %s)')
-            self.assertEqual(query.params, (60, 10))
+            self.assertEqual(tuple(query.params), (60, 10))
 
             query = self.table.select(
                 self.table.c1.as_('col1'), self.table.c2.as_('col2'),
@@ -303,7 +328,7 @@ class TestSelect(unittest.TestCase):
                         'FROM "t" AS "c") AS "b" '
                     'WHERE (ROWNUM <= %s)) AS "a" '
                 'WHERE ("rnum" > %s)')
-            self.assertEqual(query.params, (60, 10))
+            self.assertEqual(tuple(query.params), (60, 10))
 
             subquery = query.select(query.col1, query.col2)
             self.assertEqual(str(subquery),
@@ -317,7 +342,7 @@ class TestSelect(unittest.TestCase):
                     'WHERE ("rnum" > %s)) AS "a"')
             # XXX alias of query is reused but not a problem
             # as it is hidden in subquery
-            self.assertEqual(query.params, (60, 10))
+            self.assertEqual(tuple(query.params), (60, 10))
 
             query = self.table.select(limit=50, offset=10,
                 order_by=[self.table.c])
@@ -327,14 +352,14 @@ class TestSelect(unittest.TestCase):
                         'SELECT * FROM "t" AS "c" ORDER BY "c"."c") AS "b" '
                     'WHERE (ROWNUM <= %s)) AS "a" '
                 'WHERE ("rnum" > %s)')
-            self.assertEqual(query.params, (60, 10))
+            self.assertEqual(tuple(query.params), (60, 10))
 
             query = self.table.select(limit=50)
             self.assertEqual(str(query),
                 'SELECT "a".* FROM ('
                     'SELECT * FROM "t" AS "b") AS "a" '
                 'WHERE (ROWNUM <= %s)')
-            self.assertEqual(query.params, (50,))
+            self.assertEqual(tuple(query.params), (50,))
 
             query = self.table.select(offset=10)
             self.assertEqual(str(query),
@@ -342,7 +367,7 @@ class TestSelect(unittest.TestCase):
                     'SELECT "b".*, ROWNUM AS "rnum" FROM ('
                         'SELECT * FROM "t" AS "c") AS "b") AS "a" '
                 'WHERE ("rnum" > %s)')
-            self.assertEqual(query.params, (10,))
+            self.assertEqual(tuple(query.params), (10,))
 
             query = self.table.select(self.table.c.as_('col'),
                 where=self.table.c >= 20,
@@ -354,7 +379,7 @@ class TestSelect(unittest.TestCase):
                         'WHERE ("c"."c" >= %s)) AS "b" '
                     'WHERE (ROWNUM <= %s)) AS "a" '
                 'WHERE ("rnum" > %s)')
-            self.assertEqual(query.params, (20, 60, 10))
+            self.assertEqual(tuple(query.params), (20, 60, 10))
         finally:
             Flavor.set(Flavor())
 
@@ -363,14 +388,14 @@ class TestSelect(unittest.TestCase):
         query = self.table.select(c, for_=For('UPDATE'))
         self.assertEqual(str(query),
             'SELECT "a"."c" FROM "t" AS "a" FOR UPDATE')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_copy(self):
         query = self.table.select()
         copy_query = deepcopy(query)
         self.assertNotEqual(query, copy_query)
         self.assertEqual(str(copy_query), 'SELECT * FROM "t" AS "a"')
-        self.assertEqual(copy_query.params, ())
+        self.assertEqual(tuple(copy_query.params), ())
 
     def test_with(self):
         w = With(query=self.table.select(self.table.c1))
@@ -379,7 +404,7 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(query),
             'WITH "a" AS (SELECT "b"."c1" FROM "t" AS "b") '
             'SELECT * FROM "a" AS "a"')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
     def test_window(self):
         query = self.table.select(Min(self.table.c1,
@@ -388,13 +413,13 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(query),
             'SELECT MIN("a"."c1") OVER "b" FROM "t" AS "a" '
             'WINDOW "b" AS (PARTITION BY "a"."c2")')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         query = self.table.select(Rank(window=Window([])))
         self.assertEqual(str(query),
             'SELECT RANK() OVER "b" FROM "t" AS "a" '
             'WINDOW "b" AS ()')
-        self.assertEqual(query.params, ())
+        self.assertEqual(tuple(query.params), ())
 
         window = Window([self.table.c1])
         query = self.table.select(
@@ -404,7 +429,7 @@ class TestSelect(unittest.TestCase):
             'SELECT RANK() FILTER (WHERE ("a"."c1" > %s)) OVER "b", '
             'MIN("a"."c1") OVER "b" FROM "t" AS "a" '
             'WINDOW "b" AS (PARTITION BY "a"."c1")')
-        self.assertEqual(query.params, (0,))
+        self.assertEqual(tuple(query.params), (0,))
 
         window = Window([DatePart('year', self.table.date_col)])
         query = self.table.select(
@@ -412,7 +437,40 @@ class TestSelect(unittest.TestCase):
         self.assertEqual(str(query),
             'SELECT MIN("a"."c1") OVER "b" FROM "t" AS "a" '
             'WINDOW "b" AS (PARTITION BY DATE_PART(%s, "a"."date_col"))')
-        self.assertEqual(query.params, ('year',))
+        self.assertEqual(tuple(query.params), ('year',))
+
+        window = Window([self.table.c2])
+        query = self.table.select(
+            Max(self.table.c1, window=window)
+            / Min(self.table.c1, window=window))
+        self.assertEqual(str(query),
+            'SELECT (MAX("a"."c1") OVER (PARTITION BY "a"."c2") '
+            '/ MIN("a"."c1") OVER (PARTITION BY "a"."c2")) '
+            'FROM "t" AS "a"')
+        self.assertEqual(tuple(query.params), ())
+
+        window = Window([Literal(1)])
+        query = self.table.select(
+            Max(self.table.c1, window=window)
+            / Min(self.table.c1, window=window))
+        self.assertEqual(str(query),
+            'SELECT (MAX("a"."c1") OVER (PARTITION BY %s) '
+            '/ MIN("a"."c1") OVER (PARTITION BY %s)) '
+            'FROM "t" AS "a"')
+        self.assertEqual(tuple(query.params), (1, 1))
+
+        window1 = Window([self.table.c2])
+        window2 = Window([Literal(1)])
+        query = self.table.select(
+            Max(self.table.c1, window=window1)
+            / Min(self.table.c1, window=window2),
+            windows=[window1])
+        self.assertEqual(str(query),
+            'SELECT (MAX("a"."c1") OVER "b" '
+            '/ MIN("a"."c1") OVER (PARTITION BY %s)) '
+            'FROM "t" AS "a" '
+            'WINDOW "b" AS (PARTITION BY "a"."c2")')
+        self.assertEqual(tuple(query.params), (1,))
 
     def test_order_params(self):
         with_ = With(query=self.table.select(self.table.c,
@@ -425,13 +483,13 @@ class TestSelect(unittest.TestCase):
             group_by=[Literal(5)],
             order_by=[Literal(6)],
             having=Literal(7))
-        self.assertEqual(query.params, (1, 2, 3, 4, 5, 6, 7, 8))
+        self.assertEqual(tuple(query.params), (1, 2, 3, 4, 5, 6, 7, 8))
 
     def test_no_as(self):
         query = self.table.select(self.table.c)
         try:
             Flavor.set(Flavor(no_as=True))
             self.assertEqual(str(query), 'SELECT "a"."c" FROM "t" "a"')
-            self.assertEqual(query.params, ())
+            self.assertEqual(tuple(query.params), ())
         finally:
             Flavor.set(Flavor())
